/*
 * Little buffer pool handling routines.
 *
 * It's no accident that the block size is 100 bytes (event size + 1 pointer).
 *   That may need to change if the basic size of a reply or event structure
 *   changes.  Which won't happen till at least X12, and this code will
 *   probably be rewritten by then.
 */

/* Copyright (C) 1992 by Digital Equipment Corporation and Xerox Corporation */

/*	All Rights Reserved
Permission to use, copy, modify, and distribute this software and
its documentation for any purpose is hereby granted without fee,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the names of Digital and Xerox not
be used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  Digital and
Xerox make no representations about the suitability of this software
for any purpose.  It is provided "as is" without express or implied
warranty.
*/


#define NEED_EVENTS
#include "Xlibint.h"

#define BLOCK_CHUNKS	80	/* Number of blocks to get in one Xmalloc */
#define NUM_CHUNKS	512
#define VALID_BLOCK	0xA5A5A5A5

static Mutex freeListLock = NullMutex;
static _XQBlock *freeList = NULL;

/* Some statistic collecting info */

static int _fl_chunks = 0;	/* Number of chunks allocated */
static int _fl_free = 0;	/* Number of blocks currently free */
static int _fl_alloced = 0;	/* Total umber of blocks handed out */
static int _fl_freed = 0;	/* Total umber of blocks returned */

struct _fl_block
{
    _XQBlock	block;
    long	valid;
};

static struct
{
    struct _fl_block	*start;
    struct _fl_block	*end;

}	chunks[NUM_CHUNKS] = { NULL, NULL };

static _XRefillList();

void
_XInitBufferMutex()
{
    InitMutex( &freeListLock );
}

#ifdef DEBUG_BUFF
static void
check_lists()	/* checks to see if any QBlock struct has been trashed */
{
    int		i;
    struct _fl_block *p;

    for( i = 0; i < _fl_chunks; i++ )
    {
	p = chunks[i].start;
	while( p < chunks[i].end )
	{
	    if( p->valid != VALID_BLOCK )
	    {
		fprintf( stderr, "QBlock structure garbled\n" );
		exit( 1 );
	    }
	    p++;
	}
    }
}
#endif	/* DEBUG_BUFF */


_XQBlock *
_XGetBlock()
{
    _XQBlock *ret;

#ifdef DEBUG_BUFF
    check_lists();
#endif
    LockMutex( &freeListLock );

    if (!freeList)	/* Nothing on the list? */
	_XRefillList();	/* Refill it! */

    ret = freeList;

    if (freeList)
    {
	freeList = (_XQBlock*)freeList->next;
	_fl_alloced++;
	_fl_free--;
    }

    UnlockMutex( &freeListLock );
    return ret;
}

void
_XFreeBlock(block)
    _XQBlock *block;
{
#ifdef DEBUG_BUFF
    int		i;
    int		valid = False;

    if( ((struct _fl_block *)block)->valid != VALID_BLOCK )
    {
	fprintf( stderr, "attempt to free garbled QBlock!\n" );
	exit( 1 );
    }
    for( i = 0; i < _fl_chunks; i++ )
    {
	if(( block >= (_XQBlock*)chunks[i].start ) && ( block < (_XQBlock*)chunks[i].end ))
	{
	    valid = True;
	}
    }
    if( !valid )
    {
	fprintf( stderr, "attempt to free invalid QBlock!\n" );
	exit( 1 );
    }
    check_lists();
#endif

    LockMutex( &freeListLock );

    block->next = (XPointer)freeList;
    freeList = (_XQBlock *)block;

    _fl_freed++;
    _fl_free++;

    UnlockMutex( &freeListLock );
}

/*
 * Refill the free list - Do not call unless freeListLock is held, and the
 *   list is empty!
 */

#ifdef DEBUG_BUFF
static
_XRefillList()
{
    struct _fl_block	* p;
    int i;

    fprintf( stderr, "event queue free list empty, refilling\n" );

    p = (struct _fl_block *)Xmalloc( sizeof(struct _fl_block) * BLOCK_CHUNKS );
    if( ! p )
    {
	fprintf( stderr, "Not enough memory in _XRefillList\n" );
	exit( 1 );
    }

    for (i = 0; i < BLOCK_CHUNKS - 1; i++)	/* Link the list together */
    {
	p[i].block.next = (XPointer)&p[i+1].block;
	p[i].valid = VALID_BLOCK;
    }
    p[i].block.next = NULL;		/* terminate the list */
    p[i].valid = VALID_BLOCK;

    chunks[_fl_chunks].start = p;
    chunks[_fl_chunks].end = p + BLOCK_CHUNKS;

    _fl_chunks++;
    _fl_free += BLOCK_CHUNKS;

    freeList = &p->block;
}

#else	/* ! DEBUG_BUFF */

static
_XRefillList()
{
    _XQBlock * p;
    int i;

    p = (_XQBlock *)Xmalloc( sizeof(_XQBlock) * BLOCK_CHUNKS );

#ifdef NOTDEF	/* If we can figure out how to get an XIOError out of this
		    code, we should do it here.  Meanwhile, we'll just blindly
		    initialize the memory in hopes of generating a core dump.
		*/
    if (!p)		/* Didn't get memory */
	return ;
#endif /* NOTDEF */

    for (i = 0; i < BLOCK_CHUNKS - 1; i++)	/* Link the list together */
	p[i].next = (XPointer)&p[i+1];

    p[i].next = NULL;		/* terminate the list */

    _fl_chunks++;
    _fl_free += BLOCK_CHUNKS;

    freeList = p;
}
#endif	/* DEBUG_BUFF */


/*
 * Xfree - This routine checks to see if the memory being freed happens to be
 * part of the list of event blocks that are never supposed to be freed.  This
 * routine is for debugging purposes only!  To enable it, comment out the
 * #define for Xfree in PCR.h, compile this file with DEBUG_BUFF=1 and recompile the
 * remainder of Xlib as usual.
 */

#ifdef DEBUG_BUFF
void
Xfree( block )
    char	*block;
{
    int		i;
    int		valid = True;

    for( i = 0; i < _fl_chunks; i++ )
    {
	if(( block >= (char *)chunks[i].start ) && ( block < (char *)chunks[i].end ))
	{
	    valid = False;
	}
    }
    if( !valid )
    {
	fprintf( stderr, "Attempt to Xfree QBlock memory!\n" );
	exit( 1 );
    }

    check_lists();

    XR_free( block );

}	/* end of Xfree function */

#ifndef PCR
char *Xmalloc( size )
    unsigned	size;
{
    return( (char *)malloc( size ));
}

char *Xcalloc( count, size )
    unsigned	count, size;
{
    return( (char *)calloc( count, size ));
}

char *Xrealloc( ptr, size )
    char	*ptr;
    unsigned	size;
{
    return( (char *)realloc( ptr, size ));
}
#endif	/* PCR */

#endif	/* DEBUG_BUFF */
