/*
 * XlibTK.c - Toolkit for testing the MT Xlib
 */

#include	<stdlib.h>
#include	<stdio.h>
#include	<pthread.h>
#include	<X11/Xlib.h>
#include	"XlibTK.h"

/*
 * useful stuff
 */

extern long _Xevent_to_mask[];
#define EventToMask(type)	(_Xevent_to_mask[(type)])

/*
 * local routines
 */

static void TKInitLocks(
#if NeedFunctionPrototypes
    void
#endif
);
static void TKDoAdd(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    HandlerList	/* handler */
#endif
);
static void TKDoRemove(
#if NeedFunctionPrototypes
    AppContext	/* context */,
    HandlerList	/* handler */
#endif
);
static Boolean TKEventFilter(
#if NeedFunctionPrototypes
    Display *	/* dpy */,
    XQueue *	/* queue */,
    XEvent *	/* event */,
    AppContext	/* context */
#endif
);

/*
 * list of application contexts
 */

static AppContext	appContextList = NULL;
static pthread_mutex_t	contextListLock;
static pthread_once_t	once = pthread_once_init;

/*
 * start of TKInitLocks function
 */

static void
TKInitLocks()
{
    pthread_mutex_init( &contextListLock, pthread_mutexattr_default );

}	/* end of TKInitLocks function */

/*
 * start of TKInitialize function
 */

AppContext
TKInitialize( dpy_name )
    char	*dpy_name;
{
    Display	*dpy;
    AppContext	context;

    pthread_once( &once, TKInitLocks );

/*
 * open display connection
 */

    dpy = XOpenDisplay( dpy_name );
    if( dpy == NULL )
    {
	fprintf( stderr, "Failed opening display - %s.\n", dpy_name );
	exit( 1 );
    }

/*
 * malloc and initialize application context
 */

    context = (AppContext)malloc( sizeof( *context ));
    context->dpy = dpy;
    context->queue = DefaultEventQueue( dpy );
    context->event_procs = NULL;
    context->cleanup_procs = NULL;
    context->timeout_procs = NULL;
    context->windows = NULL;
    context->add_list = NULL;
    context->remove_flag = FALSE;
    context->status = AC_STOPPED;
/*
 * initialize the mutexes
 */
    pthread_mutex_init( &context->lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->event_lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->cleanup_lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->timeout_lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->window_lock, pthread_mutexattr_default );

/*
 * add context to list
 */

    pthread_mutex_lock( &contextListLock );
    if( appContextList == NULL )
    {
	appContextList = context;
    }
    else
    {
	context->next = appContextList;
	appContextList = context;
    }
    pthread_mutex_unlock( &contextListLock );
    return( context );

}	/* end of TKInitialize function */

/*
 * start of TKEventFilter function
 */

static Boolean
TKEventFilter( dpy, queue, event, context )
    Display	*dpy;
    XQueue	*queue;
    XEvent	*event;
    AppContext	context;
{
    WindowList		curr;
    unsigned long 	mask;
/*
 * search the window list to determine if this event is wanted
 */
    pthread_mutex_lock( &context->window_lock );
    for( curr = context->windows; curr != NULL; curr = curr->next )
    {
	if( curr->window == event->xany.window )
	{
	    mask = EventToMask( event->type );
	    if(( mask == 0 )	/* non-maskable */
	       || (( mask & curr->mask ) != 0 ))
	    {
		pthread_mutex_unlock( &context->window_lock );
		return( TRUE );
	    }
	    break;
	}
    }
/*
 * not found, return FALSE
 */
    pthread_mutex_unlock( &context->window_lock );
    return( FALSE );

}	/* end of TKEventFilter function */

/*
 * start of TKCreateAppContext function
 */

AppContext
TKCreateAppContext( dpy )
    Display	*dpy;
{
    AppContext	context;
    long	value_mask;

/*
 * malloc and initialize application context
 */

    context = (AppContext)malloc( sizeof( *context ));
    context->dpy = dpy;
    context->event_procs = NULL;
    context->cleanup_procs = NULL;
    context->timeout_procs = NULL;
    context->windows = NULL;
    context->add_list = NULL;
    context->remove_flag = FALSE;
    context->status = AC_STOPPED;
/*
 * initialize the mutexes
 */
    pthread_mutex_init( &context->lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->event_lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->cleanup_lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->timeout_lock, pthread_mutexattr_default );
    pthread_mutex_init( &context->window_lock, pthread_mutexattr_default );

/*
 * create event queue
 */

    context->props.mask = 0;
    context->props.predicate = TKEventFilter;
    context->props.pred_data = (void *)context;
    value_mask = QPMask | QPPredicate | QPPredicateData;
    context->queue = XCreateEventQueue( dpy, value_mask, &context->props );

/*
 * add context to list	LOCK THIS
 */

    pthread_mutex_lock( &contextListLock );
    context->next = appContextList;
    appContextList = context;
    pthread_mutex_unlock( &contextListLock );

    return( context );

}	/* end of TKCreateAppContext function */

/*
 * start of TKDestroyAppContext function
 */

Boolean
TKDestroyAppContext( context )
    AppContext	context;
{
    AppContext	curr_context;
    AppContext	prev_context = NULL;
    HandlerList	curr_handler;
    HandlerList	next_handler;
    WindowList	curr_window;
    WindowList	next_window;

/*
 * check that context is valid
 */
    pthread_mutex_lock( &context->lock );
    if( context->status != AC_STOPPED )
    {
	fprintf( stderr, "Attempt to destroy a running context\n" );
	return( FALSE );
    }
    
    pthread_mutex_lock( &contextListLock );
    curr_context = appContextList;
    while( curr_context != NULL )
    {
	if( curr_context = context )
	{
	    break;
	}
	prev_context = curr_context;
	curr_context = curr_context->next;
    }

    if( curr_context == NULL )
    {
	pthread_mutex_unlock( &contextListLock );
	return( FALSE );
    }

/*
 * remove context from list
 */

    if( prev_context == NULL )
    {
	appContextList = curr_context->next;
    }
    else
    {
	prev_context->next = curr_context->next;
    }
    pthread_mutex_unlock( &contextListLock );

/*
 * destroy event queue
 */

    XDestroyEventQueue( context->dpy, context->queue );

/*
 * free the handlers
 */

    for( curr_handler = context->event_procs; curr_handler != NULL;
	curr_handler = next_handler )
    {
	next_handler = curr_handler->next;
	free( curr_handler );
    }

    for( curr_handler = context->cleanup_procs; curr_handler != NULL;
	curr_handler = next_handler )
    {
	next_handler = curr_handler->next;
	free( curr_handler );
    }

    for( curr_handler = context->timeout_procs; curr_handler != NULL;
	curr_handler = next_handler )
    {
	next_handler = curr_handler->next;
	free( curr_handler );
    }

/*
 * free the windows
 */

    for( curr_window = context->windows; curr_window != NULL;
	curr_window = next_window )
    {
	next_window = curr_window->next;
	free( curr_window );
    }

/*
 * free the context structure and return
 */

    free( context );
    return( TRUE );

}	/* end of TKDestroyAppContext function */

/*
 * start of DpyFromContext function  -- Could be done in a macro
 */

Display *
TKDpyFromContext( context )
    AppContext	context;
{
    return( context->dpy );

}	/* end of DpyFromContext function */

/*
 * start of TKMainLoop function
 */

void
TKMainLoop( context )
    AppContext	context;
{
    HandlerList	curr;
    HandlerList	list;
    WindowList	curr_win;
    XEvent	event;
    int		status;

/*
 * check state, save thread ID and set state to running	LOCK THIS SECTION
 */

    if( context->status != AC_STOPPED )
    {
	fprintf( stderr, "Main loop allready running!\n" );
	return;
    }
    context->thread = pthread_self();
    context->status = AC_RUNNING;

/*
 * run until somebody signals a stop
 */

    while( context->status == AC_RUNNING )
    {
	XNextEventMQ( context->dpy, context->queue, &event );
/*
 * dispatch event
 */
	pthread_mutex_lock( &context->lock );
	if( context->status == AC_RUNNING )
	{
	    context->status = AC_DISPATCH;
	}
	else
	{
	    pthread_mutex_unlock( &context->lock );
	    break;
	}
	pthread_mutex_unlock( &context->lock );
/*
 * find the window first
 */
	for( curr_win = context->windows; curr_win != NULL;
	     curr_win = curr_win->next )
	{
	    if( curr_win->window == event.xany.window )
	    {
		break;
	    }
	}
/*
 * call any handlers interested in this event
 */
	if( curr_win != NULL )
	{
	    for( curr = curr_win->handlers;
		( curr != NULL ) && ( curr->window->window == event.xany.window );
		curr = curr->next )
	    {
		if(( curr->valid )
		   && ((( curr->mask & EventToMask( event.type )) != 0 )
		       || (( EventToMask( event.type ) == 0 )
			   && curr->non_maskable )))
		{
		    curr->handler( context, curr->window->window, &event,
				   curr->data );
		}
	    }
	}
/*
 * done dispatching
 */
	pthread_mutex_lock( &context->lock );
	if( context->status == AC_DISPATCH )
	{
	    context->status = AC_RUNNING;
	}
	else
	{
	    pthread_mutex_unlock( &context->lock );
	    break;
	}
	pthread_mutex_unlock( &context->lock );
/*
 * check for added or removed handlers
 */
	pthread_mutex_lock( &context->event_lock );
	if( context->add_list != NULL )
	{
	    TKDoAdd( context, context->add_list );
	    context->add_list = NULL;
	}
	if( context->remove_flag )
	{
	    TKDoRemove( context, NULL );
	    context->remove_flag = FALSE;
	}
	pthread_mutex_unlock( &context->event_lock );

    }	/* end of while RUNNING loop */

/*
 * call the cleanup procs
 */

    while(( curr = context->cleanup_procs ) != NULL )
    {
	context->cleanup_procs = curr->next;
	curr->handler( context, curr->data );
	free( curr );
    }
    context->status = AC_STOPPED;
    return;

}	/* end of TKMainLoop function */

/*
 * start of TKExitMainLoop function
 */

void
TKExitMainLoop( context )
    AppContext	context;
{
    pthread_mutex_lock( &context->lock );
    context->status = AC_SHUTDOWN;
    pthread_mutex_unlock( &context->lock );
    return;

}	/* end of TKExitMainLoop function */

/*
 * start of TKAddEventHandler function
 */

HandlerStruct *
TKAddEventHandler( context, handlerProc, clientData, win, mask )
    AppContext		context;
    void		(*handlerProc)();
    void		*clientData;
    Window		win;
    unsigned long	mask;
{
    HandlerList		handler;
    WindowList		window;

    pthread_mutex_lock( &context->lock );
    if( context->status == AC_SHUTDOWN )
    {
	pthread_mutex_unlock( &context->lock );
	return( NULL );
    }

/*
 * create the new handler struct
 */
    handler = (HandlerList)malloc( sizeof( *handler ));
    handler->handler = handlerProc;
    handler->data = clientData;
    handler->mask = mask;
    handler->valid = TRUE;
    if( mask == 0 )
    {
	handler->non_maskable = TRUE;
    }
    else
    {
	handler->non_maskable = FALSE;
    }
/*
 * find the window structure
 */
    pthread_mutex_lock( &context->window_lock );
    for( window = context->windows; window != NULL; window = window->next )
    {
	if( window->window == win )
	{
	    break;
	}
    }
    if( window == NULL )	/* new window in this context */
    {
	window = (WindowList)malloc( sizeof( *window ));
	window->window = win;
	window->mask = 0;
	window->handlers = handler;
	window->next = NULL;
    }
    handler->window = window;
    window->mask |= mask;
    pthread_mutex_unlock( &context->window_lock );
/*
 * put handler in list
 */
    pthread_mutex_lock( &context->event_lock );
    if( context->status == AC_DISPATCH )
    {
	handler->next = context->add_list;
	context->add_list = handler;
    }
    else
    {
	handler->next = NULL;
	TKDoAdd( context, handler );
    }
    pthread_mutex_unlock( &context->event_lock );
    pthread_mutex_unlock( &context->lock );

    return( handler );

}	/* end of TKAddEventHandler function */

/*
 * start of TKDoAdd function
 * assumes that event_lock is already held
 */
static void
TKDoAdd( context, handler )
    AppContext		context;
    HandlerList		handler;
{
    HandlerList		next_handler;
    WindowList		window;
    WindowList		curr_win;
    WindowList		prev_win;
    HandlerList		curr_handler;
    AppContext		curr_con;
    unsigned long	new_mask;

    while( handler != NULL )
    {
	next_handler = handler->next;
	window = handler->window;

	if( handler == window->handlers )	/* new window */
	{
	    if( context->windows == NULL )
	    {
		context->windows = window;
		context->event_procs = handler;
		handler->next = NULL;
		handler->prev = NULL;
	    }
	    else	/* put window and handler at end of lists */
	    {
		curr_win = context->windows;
		while( curr_win->next != NULL )
		{
		    curr_win = curr_win->next;
		}
		curr_win->next = window;
		curr_handler = curr_win->handlers;
		while( curr_handler->next != NULL )
		{
		    curr_handler = curr_handler->next;
		}
		curr_handler->next = handler;
		handler->prev = curr_handler;
		handler->next = NULL;
	    }
	}
	else	/* add handler to list for this window */
	{
	    curr_handler = window->handlers;
	    handler->next = curr_handler->next;
	    if( handler->next != NULL )
	    {
		handler->next->prev = handler;
	    }
	    curr_handler->next = handler;
	    handler->prev = curr_handler;
	}
	
/*
 * calculate new event mask for window and register it
 */
	
	new_mask = 0;
	pthread_mutex_lock( &contextListLock );
	for( curr_con = appContextList; curr_con != NULL;
	    curr_con = curr_con->next )
	{
	    for( curr_win = curr_con->windows; curr_win != NULL;
		curr_win = curr_win->next )
	    {
		if( curr_win->window == window->window )
		{
		    break;
		}
	    }
	    if( curr_win != NULL )
	    {
		new_mask |= curr_win->mask;
	    }
	}
	pthread_mutex_unlock( &contextListLock );
	
	XSelectInput( context->dpy, window->window, new_mask );
	
	context->props.mask |= handler->mask;
	XChangeEventQueueProperties( context->dpy, context->queue, QPMask,
				    &context->props );
	handler = next_handler;

    }	/* end of while next */
    return;

}	/* end of TKDoAdd function */

/*
 * start of TKRemoveEventHandler function
 */

Boolean
TKRemoveEventHandler( context, handler )
    AppContext		context;
    HandlerStruct	*handler;
{

    pthread_mutex_lock( &context->lock );
    if( context->status == AC_SHUTDOWN )
    {
	pthread_mutex_unlock( &context->lock );
	return( FALSE );
    }
/*
 * remove handler from list
 */
    pthread_mutex_lock( &context->event_lock );
    if( handler->valid )
    {
	handler->valid = FALSE;
    }
    else	/* already removed */
    {
	pthread_mutex_unlock( &context->event_lock );
	pthread_mutex_unlock( &context->lock );
	return( FALSE );
    }
	
    if( context->status == AC_DISPATCH )
    {
	context->remove_flag = TRUE;
    }
    else
    {
	TKDoRemove( context, handler );
    }
    pthread_mutex_unlock( &context->event_lock );
    pthread_mutex_unlock( &context->lock );

    return( TRUE );

}	/* end of TKRemoveEventHandler function */

/*
 * start of TKDoRemove function
 * assumes that event_lock is already held
 */

static void
TKDoRemove( context, handler )
    AppContext		context;
    HandlerList		handler;
{
    AppContext		curr_con;
    WindowList		window;
    WindowList		curr_win;
    HandlerList		curr_hand;
    unsigned long	new_mask;
    Window		win;
    
    if( handler == NULL )
    {
	return;
    }

    window = handler->window;
    win = window->window;
/*
 * first remove the handler from the handler list
 */
    if( handler->prev == NULL )	/* first handler in list */
    {
	context->event_procs = handler->next;
    }
    else
    {
	handler->prev->next = handler->next;
    }
/*
 * now remove the handler from the window handler list
 */
    if( window->handlers == handler )
    {
	if(( handler->next != NULL ) && ( handler->next->window == window ))
	{
	    window->handlers = handler->next;
	}
	else	/* only handler for this window, remove window */
	{
	    if( window == context->windows )	/* first window */
	    {
		context->windows = window->next;
	    }
	    else
	    {
		curr_win = context->windows;
		while( curr_win->next != NULL )
		{
		    if( curr_win->next == window )
		    {
			curr_win->next = window->next;
			break;
		    }
		    curr_win = curr_win->next;
		}
	    }
	    free( window );
	}
    }
    free( handler );
/*
 * calculate new event mask for window and register it
 */
    new_mask = 0;
    pthread_mutex_lock( &contextListLock );
    for( curr_con = appContextList; curr_con != NULL;
	curr_con = curr_con->next )
    {
	for( curr_win = curr_con->windows; curr_win != NULL;
	    curr_win = curr_win->next )
	{
	    if( curr_win->window = win )
	    {
		break;
	    }
	}
	if( curr_win != NULL )
	{
	    new_mask |= curr_win->mask;
	}
    }
    pthread_mutex_unlock( &contextListLock );
    XSelectInput( context->dpy, win, new_mask );

    new_mask = 0;
    for( curr_hand = context->event_procs; curr_hand != NULL;
	curr_hand = curr_hand->next )
    {
	new_mask |= curr_hand->mask;
    }
    
    context->props.mask = new_mask;
    XChangeEventQueueProperties( context->dpy, context->queue, QPMask,
				&context->props );
    
    return;

}	/* end of TKDoRemove function */

/*
 * start of TKAddCleanupProc function
 */

HandlerStruct *
TKAddCleanupProc( context, proc, data )
    AppContext		context;
    int			(*proc)();
    void		*data;
{
    HandlerList		handler;

    pthread_mutex_lock( &context->cleanup_lock );
    handler = (HandlerList)malloc( sizeof( *handler ));
    handler->handler = proc;
    handler->data = data;

    handler->next = context->cleanup_procs;
    context->cleanup_procs = handler;
    if( handler->next != NULL )
    {
	handler->next->prev = handler;
    }
    handler->prev = NULL;
    pthread_mutex_unlock( &context->cleanup_lock );

    return( handler );

}	/* end of TKAddCleanupProc function */

/*
 * start of TKRemoveCleanupProc function
 */

Boolean
TKRemoveCleanupProc( context, handler )
    AppContext		context;
    HandlerStruct	*handler;
{
    pthread_mutex_lock( &context->cleanup_lock );
    if( handler->next != NULL )
    {
	handler->next->prev = handler->prev;
    }
    if( handler->prev != NULL )
    {
	handler->prev->next = handler->next;
    }
    else
    {
	context->cleanup_procs = handler->next;
    }
    pthread_mutex_unlock( &context->cleanup_lock );

    return( TRUE );

}	/* end of TKRemoveCleanupProc function */

/*
 * start of TKAddTimeout function
 */

HandlerStruct *
TKAddTimeout( context, handlerProc, clientData, msecs )
    AppContext		context;
    int			(*handlerProc)();
    void		*clientData;
    long		msecs;
{

    return( NULL );

}	/* end of TKAddTimeout function */

/*
 * start of TKRemoveTimeout function
 */

Boolean
TKRemoveTimeout( context, handler )
    AppContext		context;
    HandlerStruct	*handler;
{
    return( FALSE );

}	/* end of TKRemoveTimeout function */

